#!/usr/bin/env python
"""
Integration tests for wrapper functionality in the matop submodule.

Public use of this module is not supported by NAG.
"""

# NAG Copyright 2018-2021.

# pylint: disable=invalid-name

import unittest
import warnings

import numpy as np

from test_functionality_core import IntegrationTest

from naginterfaces.base import utils as b_utils
from naginterfaces.base import matop as b_chapter
from naginterfaces.library import blas
from naginterfaces.library import matop as chapter

class TestF01cr(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f01cr wrapper.
    """

class TestF01crBase(TestF01cr, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF01cr.set_up(self)
        self.fun = b_chapter.real_gen_trans_inplace

    def test_good(self):
        """
        Good call.
        """
        m = 3
        n = 7
        a = np.array(range(1, m*n+1), dtype=float)
        self.fun(a, m, n, m*n)
        np.testing.assert_equal(
            a,
            [
                1.0, 4.0, 7.0, 10.0, 13.0, 16.0, 19.0,
                2.0, 5.0, 8.0, 11.0, 14.0, 17.0, 20.0,
                3.0, 6.0, 9.0, 12.0, 15.0, 18.0, 21.0,
            ],
        )

class TestF01df(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f01df wrapper.
    """

class TestF01dfNp(TestF01df, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF01df.set_up(self)
        self.fun = chapter.real_tri_matmul

    def test_good_no_c(self):
        """
        Good call.
        """
        c = self.fun(
            'U', 'T', 'N', 0.2,
            [
                [1.4, 0, 0, 0],
                [6.1, 8.1, 0, 0],
                [1.4, 5.9, 3.1, 0],
                [9.4, 5.1, 4.8, 1.9],
            ],
            [
                [9.2, 4.8, 1.4, 3.7],
                [0, 4.7, 1.1, 2.7],
                [0, 0, 6.1, 5.7],
                [0, 0, 0, 4.8],
            ],
            0.0,
        )
        np.testing.assert_allclose(
            c[0,:],
            [2.576, 7.078, 3.442, 14.95],
        )
        np.testing.assert_allclose(
            c[1,1:],
            [7.614, 8.98, 15.996],
        )
        np.testing.assert_allclose(
            c[2,2:],
            [3.782, 8.142],
        )
        np.testing.assert_allclose(
            c[3,3:],
            [1.824],
        )

    def test_good_c(self):
        """
        Good call.
        """
        c = self.fun(
            'U', 'T', 'N', 0.2,
            [
                [1.4, 0, 0, 0],
                [6.1, 8.1, 0, 0],
                [1.4, 5.9, 3.1, 0],
                [9.4, 5.1, 4.8, 1.9],
            ],
            [
                [9.2, 4.8, 1.4, 3.7],
                [0, 4.7, 1.1, 2.7],
                [0, 0, 6.1, 5.7],
                [0, 0, 0, 4.8],
            ],
            0.7,
            [
                [4.0, 7.8, 9.0, 3.7],
                [3.2, 3.1, 1.4, 1.2],
                [6.9, 6.2, 1.9, 6.8],
                [1.1, 4.2, 8.4, 7.1],
            ],
        )
        np.testing.assert_allclose(
            c,
            [
                [5.376, 12.538, 9.742, 17.54],
                [3.2, 9.784, 9.96, 16.836],
                [6.9, 6.2, 5.112, 12.902],
                [1.1, 4.2, 8.4, 6.794],
            ],
        )

class TestF01dt(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f01dt wrapper.
    """

class TestF01dtNp(TestF01dt, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF01dt.set_up(self)
        self.fun = chapter.complex_tri_matmul

    def test_good_no_c(self):
        """
        Good call.
        """
        c = self.fun(
            'U', 'C', 'N', 0.2 + 2.8j,
            [
                [3.3 + 2.9j, 0, 0, 0],
                [6.4 + -8.9j, 1.6 + 1.2j, 0, 0],
                [2.0 + -6.4j, 3.4 + 4.9j, 1.8 + -4.5j, 0],
                [6.1 + 1.0j, 5.0 + -4.2j, 2.4 + 0.2j, 6.7 + 3.0j],
            ],
            [
                [3.2 + -4.4j, 1.6 + -5.2j, 5.2 + 1.6j, 2.1 + 8.3j],
                [0, 2.8 + 6.8j, 1.6 + -4.0j, 4.0 + -0.8j],
                [0, 0, 3.2 + 2.5j, 1.0 + -1.5j],
                [0, 0, 0, 5.9 + -6.4j],
            ],
            0.j,
        )
        np.testing.assert_allclose(
            c[0,:],
            [
                66.2 - 10.92j, -141.072 - 137.392j,
                -0.488 + 163.376j, -7.69 + 295.796j,
            ],
        )
        np.testing.assert_allclose(
            c[1,1:],
            [
                -18.528 + 36.896j, 47.578 + 55.392j,
                76.814 + 157.376j,
            ],
        )
        np.testing.assert_allclose(
            c[2,2:],
            [-54.018 - 11.592j, 45.558 + 57.056j],
        )
        np.testing.assert_allclose(
            c[3,3:],
            [173.69 + 44.808j],
        )

    def test_good_c(self):
        """
        Good call.
        """
        c = self.fun(
            'U', 'C', 'N', 0.2 + 2.8j,
            [
                [3.3 + 2.9j, 0, 0, 0],
                [6.4 + -8.9j, 1.6 + 1.2j, 0, 0],
                [2.0 + -6.4j, 3.4 + 4.9j, 1.8 + -4.5j, 0],
                [6.1 + 1.0j, 5.0 + -4.2j, 2.4 + 0.2j, 6.7 + 3.0j],
            ],
            [
                [3.2 + -4.4j, 1.6 + -5.2j, 5.2 + 1.6j, 2.1 + 8.3j],
                [0, 2.8 + 6.8j, 1.6 + -4.0j, 4.0 + -0.8j],
                [0, 0, 3.2 + 2.5j, 1.0 + -1.5j],
                [0, 0, 0, 5.9 + -6.4j],
            ],
            0.7 + 0.5j,
            [
                [4.0 + 1.7j, 7.8 + 0.1j, 9.0 + 5.1j, 3.7 + 9.1j],
                [3.2 + 8.8j, 3.1 + 8.1j, 1.4 + 6.7j, 1.2 + 1.4j],
                [6.9 + 1.2j, 6.2 + 2.6j, 1.9 + 4.4j, 6.8 + 2.0j],
                [1.1 + 4.5j, 4.2 + 0.8j, 8.4 + 3.4j, 7.1 + 6.1j],
            ],
        )
        np.testing.assert_allclose(
            c,
            [
                [
                    68.15 - 7.73j, -135.662 - 133.422j,
                    3.262 + 171.446j, -9.65 + 304.016j,
                ],
                [
                    3.2 + 8.8j, -20.408 + 44.116j,
                    45.208 + 60.782j, 76.954 + 158.956j,
                ],
                [6.9 + 1.2j, 6.2 + 2.6j, -54.888 - 7.562j, 49.318 + 61.856j],
                [1.1 + 4.5j, 4.2 + 0.8j, 8.4 + 3.4j, 175.61 + 52.628j],
            ],
        )

class TestF01sa(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f01sa wrapper.
    """

class TestF01saNp(TestF01sa, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF01sa.set_up(self)
        self.fun = chapter.real_nmf
        self.a = [
            [8., 10., 5., 10., 5.],
            [9., 21., 7., 17., 10.],
            [12., 17., 8., 14., 6.],
            [14., 18., 9., 16., 7.],
            [13., 29., 10., 23., 13.],
            [10., 17., 7., 14., 7.],
        ]
        self.w = [
            [1.13762312, 1.18752198, 0.50625744],
            [3.05022932, 2.28974284, 1.12067648],
            [1.65246573, 2.61887843, 2.66302154],
            [0.29028784, 2.76216426, 1.39868493],
            [2.19317303, 2.90910074, 0.48745224],
            [2.08269566, 2.47982721, 1.08634255],
        ]
        self.h = [
            [2.37956027, 1.24992201, 3.43080604, 1.25784511, 1.40833516],
            [3.38788601, 1.81046744, 2.77209202, 3.48391992, 1.08437095],
            [0.45468824, 0.42640333, 0.27647956, 3.12439008, 1.3233102],
        ]
        self.exp_w = [
            [0.1611, 0.2912, 0.1945],
            [0.4777, 0.3146, 0.5818],
            [0.3179, 0.4475, 0.1284],
            [0.3098, 0.5199, 0.1631],
            [0.6549, 0.4609, 0.7068],
            [0.3464, 0.3651, 0.2847],
        ]
        self.exp_h = [
            [6.2177E-16, 3.0274E+01, 4.2556E+00, 6.5413E+00, 2.5214E+00],
            [2.6573E+01, 1.5723E+01, 1.4575E+01, 2.3361E+01, 8.6885E+00],
            [1.0847E+00, 2.7314E+00, 6.7700E-01, 1.1231E+01, 1.0406E+01],
        ]

    def test_good(self):
        """
        Good call.
        """
        k = 3
        warnings.simplefilter('ignore', b_utils.NagAlgorithmicWarning)
        soln = self.fun(self.a, k=k)
        warnings.simplefilter('default', b_utils.NagAlgorithmicWarning)
        np.testing.assert_allclose(
            soln.w,
            self.exp_w,
            rtol=5e-3,
        )
        np.testing.assert_allclose(
            soln.h,
            self.exp_h,
            rtol=1e-4,
        )

    def test_good_iterates(self):
        """
        Good call, initial iterates supplied.
        """
        warnings.simplefilter('ignore', b_utils.NagAlgorithmicWarning)
        soln = self.fun(self.a, w=self.w, h=self.h)
        warnings.simplefilter('default', b_utils.NagAlgorithmicWarning)
        np.testing.assert_allclose(
            soln.w,
            self.exp_w,
            rtol=5e-3,
        )
        np.testing.assert_allclose(
            soln.h,
            self.exp_h,
            rtol=1e-4,
        )

    def test_bad_k(self):
        """
        Bad call, k None.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'h or k must be provided',
            self.fun, self.a,
        )

    def test_bad_seed(self):
        """
        Bad call.
        """
        self.my_assert_raises_regex(
            b_utils.NagTypeError,
            'argument w.*must not be None when seed <= 0',
            self.fun, self.a, k=3, seed=-1,
        )

    def test_bad_iterparity(self):
        """
        Bad call, w None.
        """
        self.my_assert_raises_regex(
            b_utils.NagValueError,
            'Both or neither of w and h must be provided',
            self.fun, self.a, h=self.h, seed=-42,
        )

class TestF01sb(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f01sb wrapper.
    """

class TestF01sbNp(TestF01sb, unittest.TestCase):
    """
    Concrete level-2 tests.
    """

    def setUp(self):
        TestF01sb.set_up(self)
        self.fun = chapter.real_nmf_rcomm
        self.a = [
            [8., 10., 5., 10., 5.],
            [9., 21., 7., 17., 10.],
            [12., 17., 8., 14., 6.],
            [14., 18., 9., 16., 7.],
            [13., 29., 10., 23., 13.],
            [10., 17., 7., 14., 7.],
        ]

    def test_good(self):
        """
        Good call.
        """
        a = np.array(self.a)
        k = 3
        w = np.empty((a.shape[0], k)) # pylint: disable=unsubscriptable-object
        h = np.empty((k, a.shape[1])) # pylint: disable=unsubscriptable-object
        ht = np.empty((a.shape[1], k)) # pylint: disable=unsubscriptable-object
        comm = {}
        irevcm = 0
        warnings.simplefilter('ignore', b_utils.NagAlgorithmicWarning)
        while True:
            irevcm = self.fun(irevcm, w, h, ht, comm)
            if irevcm == 0:
                break
            if irevcm == 1:
                continue
            if irevcm == 2:
                ht = blas.dgemm(
                    'N', 'N', 1., np.ascontiguousarray(a.transpose()), w, 0.,
                )
                continue
            if irevcm == 3:
                w = blas.dgemm('N', 'N', 1., a, ht, 0.)
                continue
            raise ValueError(
                'NYI for irevcm ' + str(irevcm)
            ) # pragma: no cover

        warnings.simplefilter('default', b_utils.NagAlgorithmicWarning)
        np.testing.assert_allclose(
            w,
            [
                [0.159505, 0.29356, 0.190679],
                [0.478135, 0.312574, 0.583245],
                [0.317544, 0.448035, 0.125572],
                [0.308465, 0.522147, 0.158176],
                [0.655799, 0.457909, 0.708717],
                [0.346248, 0.364783, 0.283547],
            ],
            rtol=5e-3,
        )
        np.testing.assert_allclose(
            h,
            [
                [6.215108e-16, 2.997519e+01,
                 5.273596e+00, 7.212119e+00, 1.938524e+00],
                [2.642846e+01, 1.587965e+01,
                 1.408101e+01, 2.309246e+01, 9.011448e+00],
                [1.267314e+00, 2.921997e+00,
                 1.322868e-01, 1.085913e+01, 1.072685e+01],
            ],
            rtol=1e-4,
        )

class TestF01vm(IntegrationTest): # pylint: disable=too-few-public-methods
    """
    Abstract suite of integration tests for the f01vm wrapper.
    """

class TestF01vmBase(TestF01vm, unittest.TestCase):
    """
    Concrete base tests.
    """

    def setUp(self):
        TestF01vm.set_up(self)
        self.fun = b_chapter.ztrttp

    def test_good(self):
        """
        Good call.
        """
        a = np.array([
            [1.1 + 1.1j, 1.1 + 1.2j, 1.3 + 1.3j, 1.4 + 1.4j],
            [0., 2.2 + 2.2j, 2.3 + 2.3j, 2.4 + 2.4j],
            [0., 0., 3.3 + 3.3j, 3.4 + 3.4j],
            [0., 0., 0., 4.4 + 4.4j],
        ])
        n = a.shape[0]
        ap = np.empty(n*(n+1)//2, dtype=complex)
        self.fun('U', n, a, ap)
        np.testing.assert_allclose(
            ap,
            [
                1.1 + 1.1j, 1.1 + 1.2j, 2.2 + 2.2j, 1.3 + 1.3j, 2.3 + 2.3j,
                3.3 + 3.3j, 1.4 + 1.4j, 2.4 + 2.4j, 3.4 + 3.4j, 4.4 + 4.4j,
            ],
        )

if __name__ == '__main__': # pragma: no cover
    unittest.main()
